/*
 * Copyright (c) AT&T Corp. 1994, 1995.
 * This code is licensed by AT&T Corp.  For the
 * terms and conditions of the license, see
 * http://www.research.att.com/orgs/ssr/book/reuse
 */

/*
 * tkgen.c generate canvas commands to display a graph
 */

#include		"dotgraph.h"

#define		NONE		0
#define		NODE		1
#define		EDGE		2
#define		CLST		3

/* font modifiers */
#define REGULAR 0
#define BOLD	1
#define ITALIC  2
 
/* patterns */
#define P_SOLID 0
#define P_NONE  15
#define P_DOTTED 4			 /* i wasn't sure about this */
#define P_DASHED 11			 /* or this */

/* bold line constant */
#define WIDTH_NORMAL 1
#define WIDTH_BOLD 3

/* highlight flag */
#define HIGHLIGHT 1
#define NOHIGHLIGHT 0
 
#define ANONCHAR		'_'
#define ISEMPTYSTR(s)	(((s) == NULL) || (*(s) == '\0'))
#define NULL_FN(t)		(t(*)())0

static FILE		*Outfile;
static int		Obj, N_pages;
static unsigned long	ObjHandle;
static point	Pages;
static double   Scale;
static int		Rot;
static box		PB;
static int		onetime = TRUE;
static char		buffer[256];
#if TCL_MAJOR_VERSION > 7
static char		fontname[SMALLBUF] = "Times 14 normal";
#else
static char		fontname[SMALLBUF] = "-Adobe-Times-Medium-R-Normal--*-140-*";
#endif

typedef struct context_t {
	char		color[SMALLBUF], *fontfam, fontopt, font_was_set;
	char		pen, fill, penwidth;
	float		fontsz;
} context_t;

#define MAXNEST 4
static context_t cstk[MAXNEST];
static int		SP;

static char		*op[] = {"graph", "node", "edge", "graph"};

static void tkgen_append_string(char *string)
{
	Tcl_AppendElement(((tkgendata_t *)Outfile)->interp,string);
}

static void tkgen_start_item(char *item)
{
	Tcl_AppendResult(((tkgendata_t *)Outfile)->interp,
		((tkgendata_t *)Outfile)->canvas, (char *)NULL);
	tkgen_append_string("create");
	tkgen_append_string(item);
}

static void tkgen_end_item()
{
	Tcl_AppendResult(((tkgendata_t *)Outfile)->interp,"\n", (char *)NULL);
}

static void tkgen_append_point(pointf rv)
{
	sprintf(buffer, "%.1fp", rv.x);
	tkgen_append_string(buffer);
	sprintf(buffer, "%.1fp", rv.y);
	tkgen_append_string(buffer);
}

static void tkgen_append_attribute(char *attribute, char *value)
{
	tkgen_append_string(attribute);
	tkgen_append_string(value);
}

static void tkgen_append_tag(int highlight)
{
	sprintf(buffer, "%d%s%ld", highlight, op[Obj], ObjHandle);
	tkgen_append_attribute("-tags",buffer);
}

static void init_tk(void)
{
	SP = 0;
	cstk[0].color[0] = '\0';
	cstk[0].fontfam = "Times";
	cstk[0].fontopt = 0;
}

static void tkpt(point p)
{
	pointf		  tmp, rv;
	tmp.x = p.x * Scale;
	tmp.y = Scale * p.y;
	if (Rot == 0) {
		rv.x = tmp.x;
		rv.y = PB.UR.y - PB.LL.y - tmp.y;
	} else {
		rv.x = PB.UR.x - PB.LL.x - tmp.y;
		rv.y = tmp.x;
	}
	tkgen_append_point(rv);
}

static void tkptarray(point *A, int n)
{
	int			 i;

	for (i = 0; i < n; i++)
		tkpt(A[i]);
}

static void tk_font(context_t *cp)
{

#if TCL_MAJOR_VERSION > 7
	char		   *style;

	style = "normal";
	switch (cp->fontopt) {
		case BOLD:
			style = "bold";
			break;
		case ITALIC:
			style = "italic";
			break;
	}
	sprintf(fontname, "%s %d %s", cp->fontfam, round(cp->fontsz), style);
#else
	char	   *fw, *fa;

	fw = "Medium";
	fa = "R";
	switch (cp->fontopt) {
		case BOLD:
			fw = "Bold";
			break;
		case ITALIC:
			fa = "I";
			break;
	}
	sprintf(fontname, "-Adobe-%s-%s-%s-Normal--*-%d-*",
		cp->fontfam, fw, fa, round(10. * cp->fontsz));
#endif
}

static void tk_reset(void)
{
	onetime = TRUE;
}

static void
tk_begin_job(FILE *ofp, graph_t *g, char **lib, char *user, char *info[], point pages)
{
	Outfile = ofp;
	Pages = pages;
	N_pages = pages.x * pages.y;
}

static void tk_end_job(void) { }

static void tk_begin_graph(graph_t *g, box bb, point pb)
{
	Agsym_t		*a;

	g = g;
	PB = bb;

	sprintf(buffer,"%dp %dp %dp %dp",0,0,PB.UR.x-PB.LL.x+1,PB.UR.y-PB.LL.y+1);
	if (!(a = agfindattr(g, "bb"))) {
		a = agraphattr(g, "bb", "");
	}
	agxset(g, a->index, buffer);

	Tcl_AppendResult(((tkgendata_t *)Outfile)->interp,
		"global __tkgen_smooth_type\n",
		"if {![info exists __tkgen_smooth_type]} {\n",
		"    if {[catch {package require Tkspline}]} {\n",
		"        puts stderr {Warning: Tkspline package not available. Edges will not be optimally smoothed.}\n",
		"        set __tkgen_smooth_type true\n"
		"    } {\n",
		"        set __tkgen_smooth_type spline\n"
		"    }\n",
		"}\n",
		(char *)NULL);

	if (onetime) {
		init_tk();
		onetime = FALSE;
	}
}

static void tk_end_graph(void) { }

static void tk_begin_page(point page, double scale, int rot, point offset)
{
	int			 page_number;
	point		   sz;

	Scale = scale;
	Rot = rot;
	page_number = page.x + page.y * Pages.x + 1;
	sz = sub_points(PB.UR, PB.LL);
}

static void tk_end_page(void) { }

static void tk_begin_cluster(graph_t *g)
{
	Obj = CLST;
	ObjHandle = g->handle;
}

static  void tk_end_cluster(void)
{
	Obj = NONE;
}

static void tk_begin_nodes(void)
{
	Obj = NONE;
}

static void tk_end_nodes(void)
{
	Obj = NONE;
}

static void tk_begin_edges(void)
{
	Obj = NONE;
}

static void tk_end_edges(void)
{
	Obj = NONE;
}

static void tk_begin_node(node_t *n)
{
	Obj = NODE;
	ObjHandle = n->handle;
}

static  void tk_end_node(void)
{
	Obj = NONE;
}

static void tk_begin_edge(edge_t *e)
{
	Obj = EDGE;
	ObjHandle = e->handle;
}

static void tk_end_edge(void)
{
	Obj = NONE;
}

static void tk_begin_context(void)
{
	/*assert(SP + 1 < MAXNEST);*/		/* unknown problem with SGI cc */
	cstk[SP + 1] = cstk[SP];
	SP++;
}

static void tk_end_context(void)
{
	int			 psp = SP - 1;
	/* assert(SP > 0); */						/* same as above */
	if (cstk[SP].font_was_set)
		tk_font(&(cstk[psp]));
	SP = psp;
}

static void tk_set_font(char *name, double size)
{
	char		   *p, *q, buf[SMALLBUF];
	context_t	  *cp;

	cp = &(cstk[SP]);
	cp->font_was_set = TRUE;
	cp->fontsz = Scale * size;
	p = strdup(name);
	if ((q = strchr(p, '-'))) {
		*q++ = 0;
		canoncolor(q, buf);   /* makes lower case alpha */
		if (strcmp(buf, "italic") == 0) {
			cp->fontopt = ITALIC;
		} else if (strcmp(q, "bold") == 0) {
			cp->fontopt = BOLD;
		}
	}
	cp->fontfam = p;
	tk_font(&cstk[SP]);
}

#if 0
static void tk_arrowhead(point p, double theta, double scale,int flag)
{
	point		   A[2];
	double		  v;
	v = cos(RADIANS(theta)) * 10.0;
	A[0].x = round(v) + p.x;
	v = sin(RADIANS(theta)) * 10.0;
	A[0].y = round(v) + p.y;
	A[1] = p;
	if (cstk[SP].pen != P_NONE) {
		tkgen_start_item("line");
		tkptarray(A, 2);
		if (cstk[SP].color[0])
			tkgen_append_attribute("-fill",cstk[SP].color);
		if (cstk[SP].penwidth == WIDTH_BOLD)
			tkgen_append_attribute("-width","2");
		if (cstk[SP].pen == P_DOTTED)
			tkgen_append_attribute("-dash","2");
		if (cstk[SP].pen == P_DASHED)
			tkgen_append_attribute("-dash","5");
		tkgen_append_attribute("-arrow","last");
		tkgen_append_tag(HIGHLIGHT);
		tkgen_end_item();
	}
}
#endif

static void tk_set_color(char *name)
{
	canoncolor(name, cstk[SP].color);
}

static void tk_set_style(char **s)
{
	char		   *line;
	context_t	  *cp;
 
	cp = &(cstk[SP]);
	while ((line = *s++)) {
		if (streq(line, "solid")) { }	 /* no-op */
		else if (streq(line, "dashed"))
			cp->pen = P_DASHED;
		else if (streq(line, "dotted"))
			cp->pen = P_DOTTED;
		else if (streq(line, "bold"))
			cp->penwidth = WIDTH_BOLD;
		else if (streq(line, "invis"))
			cp->pen = P_NONE;
		else if (streq(line, "filled"))
			cp->fill = P_SOLID;
		else if (streq(line, "unfilled")) { }	 /* no-op */
		else
			fprintf(stderr, "tk_set_style: unsupported style %s - ignoring\n", line);
	}
}

static void
tk_textline(point p, char *str, int width, double fontsz, double align)
{
	char			*astr;

	if (align == -.5) {
		astr = "center";
	} else {
		if (align < 0) {
			astr = "right";
		} else {
			astr = "left";
		}
	}
	tkgen_start_item("text");
	tkpt(p);
	tkgen_append_attribute("-justify",astr);
	tkgen_append_attribute("-text",str);
	if (cstk[SP].color[0])
		tkgen_append_attribute("-fill",cstk[SP].color);
	tkgen_append_attribute("-font",fontname);
	tkgen_append_tag(NOHIGHLIGHT);
	tkgen_end_item();
}

static void
tk_bezier(point *A, int n, int arrow_at_start, int arrow_at_end)
{
	if (cstk[SP].pen != P_NONE) {
		tkgen_start_item("line");
		tkptarray(A, n);
		if (cstk[SP].color[0])
			tkgen_append_attribute("-fill",cstk[SP].color);
		if (cstk[SP].penwidth == WIDTH_BOLD)
			tkgen_append_attribute("-width","2");
		if (cstk[SP].pen == P_DOTTED)
			tkgen_append_attribute("-dash","2");
		if (cstk[SP].pen == P_DASHED)
			tkgen_append_attribute("-dash","5");
	    Tcl_AppendResult(((tkgendata_t *)Outfile)->interp,
		  " -smooth $__tkgen_smooth_type",
		  (char *)NULL);
		tkgen_append_tag(HIGHLIGHT);
		tkgen_end_item();
	}
}

static void
tk_polygon(point *A, int n, int filled)
{
	if (cstk[SP].pen != P_NONE) {
		tkgen_start_item("polygon");
		tkptarray(A, n);
		if (filled && cstk[SP].color[0])
			tkgen_append_attribute("-fill",cstk[SP].color);
		else  /* tk polygons default to filled, so must turn off */
			tkgen_append_attribute("-fill","");
		if (cstk[SP].color[0])
			tkgen_append_attribute("-outline",cstk[SP].color);
		if (cstk[SP].penwidth == WIDTH_BOLD)
			tkgen_append_attribute("-width","2");
		if (cstk[SP].pen == P_DOTTED)
			tkgen_append_attribute("-dash","2");
		if (cstk[SP].pen == P_DASHED)
			tkgen_append_attribute("-dash","5");
			tkgen_append_tag(HIGHLIGHT);
		tkgen_end_item();
	}
}

static void
tk_ellipse(point p, int rx, int ry,int filled)
{
	point		   A[2];

	if (cstk[SP].pen != P_NONE) {
		A[0].x = p.x - rx;
		A[0].y = p.y - ry;
		A[1].x = p.x + rx;
		A[1].y = p.y + ry;
		tkgen_start_item("oval");
		tkptarray(A, 2);
		if (filled && cstk[SP].color[0])
			tkgen_append_attribute("-fill",cstk[SP].color);
		if (cstk[SP].color[0])
			tkgen_append_attribute("-outline",cstk[SP].color);
		if (cstk[SP].penwidth == WIDTH_BOLD)
			tkgen_append_attribute("-width","2");
		if (cstk[SP].pen == P_DOTTED)
			tkgen_append_attribute("-dash","2");
		if (cstk[SP].pen == P_DASHED)
			tkgen_append_attribute("-dash","5");
		tkgen_append_tag(HIGHLIGHT);
		tkgen_end_item();
	}
}

static void tk_polyline(point *A, int n)
{
	if (cstk[SP].pen != P_NONE) {
		tkgen_start_item("line");
		tkptarray(A, n);
		if (cstk[SP].color[0])
			tkgen_append_attribute("-fill",cstk[SP].color);
		if (cstk[SP].penwidth == WIDTH_BOLD)
			tkgen_append_attribute("-width","2");
		if (cstk[SP].pen == P_DOTTED)
			tkgen_append_attribute("-dash","2");
		if (cstk[SP].pen == P_DASHED)
			tkgen_append_attribute("-dash","5");
		tkgen_append_tag(HIGHLIGHT);
		tkgen_end_item();
	}
}

static void
tk_user_shape(char *name, point *A, int n, int filled)
{
	static boolean  onetime = TRUE;
	if (onetime) {
		fprintf(stderr, "custom shapes not available with this driver\n");
		onetime = FALSE;
	}
	tk_polygon(A, n, filled);
}

static void tk_comment (void *obj, attrsym_t *sym) { }

static point tk_textsize (char *str, char *fontname, double fontsz)
{
		point size = {0,0};

		return size;
}



codegen_t	   TK_CodeGen = {
	tk_reset,
	tk_begin_job, tk_end_job,
	tk_begin_graph, tk_end_graph,
	tk_begin_page, tk_end_page,
	tk_begin_cluster, tk_end_cluster,
	tk_begin_nodes, tk_end_nodes,
	tk_begin_edges, tk_end_edges,
	tk_begin_node, tk_end_node,
	tk_begin_edge, tk_end_edge,
	tk_begin_context, tk_end_context,
	tk_set_font, tk_textline,
	tk_set_color, tk_set_style,
	tk_ellipse, tk_polygon,
	tk_bezier,
	tk_polyline,
	/*	tk_arrowhead, */
	/* canvas doesn't have the rich set of arrows now supported by emit.c */
	FALSE /* bezier_has_arrows = FALSE */,
	tk_user_shape,
	tk_comment,
	tk_textsize,
};

